import logging

import genanki
from genanki import Model
from pprint import pprint

logger = logging.getLogger('autoanki')
logger.setLevel(logging.INFO)

# TODO Add this somewhere
# # All the resources for finding the
#         CHINESE_CARD_MODEL = Model(
#             1559383145,
#             'Chinese Card (autoanki)',
#             fields=[
#                 {
#                     'name': 'word',
#                     'font': 'Arial',
#                 },
#                 {
#                     'name': 'word_traditional',
#                     'font': 'Arial',
#                 },
#                 {
#                     'name': 'pinyin',
#                     'font': 'Arial',
#                 },
#                 {
#                     'name': 'definition',
#                     'font': 'Arial',
#                 },
#             ],
#             templates=[
#                 {
#                     'name': 'Card 1',
#                     'qfmt': '{{word}}\n[{{word_traditional}}]',
#                     'afmt': '{{FrontSide}}\n\n<hr id=answer>\n{{pinyin}}<br>\n{{definition}}',
#                 },
#                 # {
#                 # 'name': 'Card 2',
#                 # 'qfmt': '{{FrontSide}}\n\n<hr id=answer>\n{{Pinyin}}\n{{Back}}',
#                 # 'afmt': '{{Front}}',
#                 # },
#             ],
#             css='.card {\n font-family: arial;\n font-size: 30px;\n text-align: center;\n color: black;\n background-color: white;\n}\n',
#         )
#         my_note = genanki.Note(
#             model=CHINESE_CARD_MODEL,
#             fields=['Capital of Canada', 'Ottawa'],
#             sort_field=1
#         )
#         my_deck = genanki.Deck(
#             2023480110,
#             'Country Capitals'
#         )
#
#         my_deck.add_note(my_note)
#         genanki.Package(my_deck).write_to_file('output.apkg')

CHINESE_CARD_MODEL = Model(
    1559383145,
    'Chinese Card (autoanki)',
    fields=[
        {
            'name': 'word',
            'font': 'Arial',
        },
        {
            'name': 'word_traditional',
            'font': 'Arial',
        },
        {
            'name': 'pinyin',
            'font': 'Arial',
        },
        {
            'name': 'definition',
            'font': 'Arial',
        },
    ],
    templates=[
        {
            'name': 'Card 1',
            'qfmt': '{{word}}\n[{{word_traditional}}]',
            'afmt': '{{FrontSide}}\n\n<hr id=answer>\n{{pinyin}}<br>\n{{definition}}',
        },
        # {
        # 'name': 'Card 2',
        # 'qfmt': '{{FrontSide}}\n\n<hr id=answer>\n{{Pinyin}}\n{{Back}}',
        # 'afmt': '{{Front}}',
        # },
    ],
    css='.card {\n font-family: arial;\n font-size: 30px;\n text-align: center;\n color: black;\n background-color: white;\n}\n',
)


def _create_test_deck():
    """
    Used for testing
    :return:
    """

    print("Generating card")
    my_note = genanki.Note(
        model=CHINESE_CARD_MODEL,
        fields=['Capital of Canada', 'Ottawa'],
        sort_field=1
    )
    my_deck = genanki.Deck(
        2023480110,
        'Country Capitals'
    )
    my_deck.add_note(my_note)
    genanki.Package(my_deck).write_to_file('output.apkg')


class DeckManager:
    """
    The class to make anki decks. Create the file using generate_deck_file()
    One of the most important concepts is the id. No matter what deck the word is in, it should have the same id so the
    same card in different decks can be remembered.
    This class makes extensive use of genanki, so understanding how genanki works
        is pretty significant for understanding this.
    """

    def __init__(self):
        self.deck = genanki.Deck(
            2020000110,
            "autoankiTesting"
        )
        self.include_pinyin_numbers = False
        self.include_number_of_strokes = False

        self.book_list = []

    def generate_deck_file(self, words, deck_name: str, filename: str):
        """
        Generates a deck file from the database
        :param deck_name: The name of the deck to be created
        :param definitions_filename: The name of the file containing the definitions.
        :return:
        """
        # TODO This is a legacy function. Update this to use the database, rather than text file
        # Has not been changed to use the object, which it should.

        # Number of valid cards that have been added to the deck
        num_of_valid_cards_added = 0

        # length = general_functions.file_len(definitions_filename)
        # self.deck.add_note()
        self.deck = genanki.Deck(
            2020000110,
            deck_name
        )

        for row in words:
            word = row["word"]
            word_traditional = row["word_traditional"]
            if word_traditional == 'Same':
                word_traditional = "-"
            pinyin = row["pinyin"]
            definition = "<br>" + row["definition"]
            # word["word_id"]
            # word["word"]
            # word["word_traditional"]
            # word["word_type"]
            # word["pinyin"]
            # word["pinyin_numbers"]
            # word["number_of_strokes"]
            # word["sub_components"]
            # word["frequency"]
            # word["hsk_level"]
            # word["top_level"]]
            # word["audio_path"]
            # word["image_path"]
            # word["definition"]

            if not word:
                word = "Not found"
            if not word_traditional:
                word_traditional = "Not found"
            if not pinyin:
                pinyin = "Not found"
            if not definition:
                definition = "Not found"

            note = genanki.Note(
                model=CHINESE_CARD_MODEL,
                fields=[word, word_traditional, pinyin, definition],
                # sort_field can be used to sort when the cards appear.
                # By default they are shown in the order they are addeed, so this is not currently used
                sort_field=1,

            )
            self.deck.add_note(note)

        genanki.Package(self.deck).write_to_file(filename + ".apkg")
        logger.info("Deck " + deck_name + " created with " + str(num_of_valid_cards_added) + " cards")
        return filename + ".apkg"


if __name__ == '__main__':
    maker = DeckManager()
    maker.generate_deck_file('autoanki.apkg', 'example.txt')



from abc import ABC, abstractmethod


class Dictionary(ABC):

    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def find_word(self):
        pass

    @abstractmethod
    def size(self):
        pass



import logging
import os
# import shutil
# import glob
# from pathlib import Path
#
# from selenium.webdriver import ActionChains
# from selenium.webdriver.common.by import By
#
# from selenium import webdriver
# from os.path import isfile, join
# import time
# MAX_TXT_TO_PINYIN_SIZE = 270000

CLEANED_FILES_DIRECTORY = 'cleaned_files'
CLEANED_FILES_SUFFIX = '_cleaned'
# Sentences That should not be added to the completed file.
GARBAGE_SENTENCES = ['',
                     "",
                     "。",
                     "\n"]

logger = logging.getLogger('autoanki')
logger.setLevel(logging.INFO)


class BookCleaner:

    def __init__(self):
        self.file_list = []
        self.bookpath = ""

    def clean(self, bookpath: str):
        """
        Cleans the files contained in the bookpath. If bookpath is a single file, clean it.
        :return: The path to the cleaned file(s)
        """
        if not os.path.exists(bookpath):
            # PRACTICE Should I raise an error here, or let the caller raise one?
            logger.warning("BookCleaner: Cannot find path [" + str(bookpath) + "]")
            return None

        # If the bookpath is a single file, clean and return it
        if os.path.isfile(bookpath):
            cleaned_path = self._clean_file(bookpath, cleaned_files_root=None)
            return cleaned_path
        # Otherwise, get a list of files to clean, and return the cleaned files directory
        else:
            dirty_files = []
            for root, dirs, files in os.walk(bookpath):
                for file in files:
                    # Only clean files that are not in cleaned_files directory
                    in_cleaned = str(root).find(CLEANED_FILES_DIRECTORY) != -1
                    if not in_cleaned:
                        if '.txt' in file:
                            dirty_files.append(os.path.join(root, file))

            # Check this cleaning won't be mean to the CPU
            if len(dirty_files) > 50:
                yn = input("The number of files is very large. Are you sure you want to convert this many files? (Y/N)")
                if yn.lower() != 'y':
                    return None

        # Now we have a list of files to convert in a list
        cleaned_files = []

        # Create directory for files
        cleaned_files_root = os.path.join(bookpath, CLEANED_FILES_DIRECTORY)
        if not os.path.exists(cleaned_files_root):
            os.mkdir(cleaned_files_root)

        for file in dirty_files:
            cleaned_filepath = self._clean_file(file, cleaned_files_root=cleaned_files_root)
            cleaned_files.append(cleaned_filepath)

        # print(dirty_files)
        # print(cleaned_files)

        return cleaned_files

    @staticmethod
    def _clean_file(filepath, cleaned_files_root):
        """
        Takes a txt file and cleans it up, putting every sentence on a new line
        :param filepath: The txt file to clean
        :param cleaned_files_root: The root
        :return:
        """

        # Set the directory where the cleaned file will go
        if not cleaned_files_root:
            # root/test1.txt -> root/hello_cleaned.txt
            new_filepath = os.path.splitext(filepath)[0] + CLEANED_FILES_SUFFIX + os.path.splitext(filepath)[1]
        else:
            new_filepath = os.path.join(cleaned_files_root, '/'.join(filepath.split('/')[1:]))

        # Clean page file of characters that may cause issues.
        page_file = open(filepath, encoding='utf-8')
        page_sentences = page_file.read().split("。")
        cleaned_file = open(new_filepath, "w", encoding='utf-8')

        for page_sentence in page_sentences:
            # Clean string
            page_sentence = page_sentence\
                .lstrip()\
                .rstrip()\
                .replace("  ", " ")\
                .lstrip("\"")\
                .rstrip("\"")\
                .replace("”","'")\
                .replace("　", "")\
                .replace("“","")\
                .rstrip("。")\
                .strip("'")\
                .strip("'")
            if page_sentence not in GARBAGE_SENTENCES:
                cleaned_file.write(page_sentence + "。" + "\n")

        return new_filepath

    # @staticmethod
    # def _compact_file(page_path):
    #
    #     print("Compacting...")
    #     """
    #     # Compacts all of the different pages in this directory into a specified size.
    #     # Compacted in this context means that the files are re-organized by size, not chapter/webpage
    #     # Some websites used have a free limit of 300kb, for example. Rather then pass all of these pages through one by
    #     # one, it is more efficient to compact these into 299kb files and send those.
    #     Currently not in use
    #     # :return: True if success
    #     """
    #     all_pages_text = ""
    #
    #     for page_path in self.pages:
    #         page = open(self.bookpath + "\\" + page_path, "r", encoding="utf-8")
    #         for i in range(file_len(self.bookpath + "\\" + page_path)):
    #             all_pages_text += page.readline()
    #
    #     print("Found " + str(len(self.pages)) + " pages totalling " + str(len(all_pages_text)) + " characters.")
    #
    #     # Splits the all_pages_text into smaller parts and saves it to the compacted_pages directory.
    #     i, current_compacted_filename_number = 0, 0
    #     all_pages_sentences = all_pages_text.split("\n")
    #     current_compacted_file_size = 0
    #     current_compacted_file_text = ""
    #     while i < len(all_pages_sentences):
    #
    #         current_sentence = all_pages_sentences[i]
    #
    #         # If the current file has hit its limit, start a new compacted file and reset all variables
    #         if (len(current_sentence) + current_compacted_file_size) > self.MAX_TXT_TO_PINYIN_SIZE:
    #             current_compacted_filename_number += 1
    #             current_file = open(self.compacted_pages_directory + "\\" + "compacted-" + str(
    #                 current_compacted_filename_number) + ".txt", "w",
    #                                 encoding="utf-8")
    #             for line in current_compacted_file_text.split("\n"):
    #                 current_file.write(line + "\n")
    #             current_file.close()
    #             current_compacted_file_size = 0
    #             current_compacted_file_text = ""
    #
    #         # Chinese characters are utf-8, meaning they are 8 bytes
    #         current_compacted_file_size += len(current_sentence) * 3
    #         current_compacted_file_text += current_sentence
    #         # A newline character is one byte
    #         current_compacted_file_size += 1
    #         current_compacted_file_text += "\n"
    #
    #         i += 1
    #
    #     current_compacted_filename_number += 1
    #     current_file = open(
    #         self.compacted_pages_directory + "\\" + "compacted-" + str(current_compacted_filename_number) + ".txt",
    #         "w",
    #         encoding="utf-8")
    #     for line in current_compacted_file_text.split("\n"):
    #         current_file.write(line + "\n")
    #     current_file.close()
    #
    #     print("Done compacting")
    #     return True


import logging

from .BookCleaner import BookCleaner
from .DatabaseManager import DatabaseManager
from .Dictionary import YellowBridgeDictionary
from .DeckManager import DeckManager

__author__ = "Jarvis Coghlin"
__all__ = ['AutoAnki']
__version__ = "1.0.0"
__status__ = "Development"

logger = logging.getLogger('autoanki')
logger.setLevel(logging.INFO)
logging.basicConfig(
    # filename='HISTORYlistener.log',
    level=logging.DEBUG,
    format='%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
)

class AutoAnki:

    def __init__(self, database_filepath='autoanki.db'):
        """
        Creates an instance of autoanki.
        This creates a book cleaner, database connection, and deck maker
        :param database_filepath: The filepath for the database
        """
        logger.info("autoanki: Connecting to database...")

        self.database_filepath = database_filepath

        self.book_cleaner = BookCleaner()
        if not DatabaseManager.is_database(database_filepath):
            logger.info("Creating database...")
            DatabaseManager.create_autoanki_db(database_filepath)
            logger.info("Done creating database.")
        self.database_manager = DatabaseManager(database_filepath)
        self.dictionary = YellowBridgeDictionary()
        self.deck_manager = DeckManager()

        logger.info("autoanki: Connected!")

    def add_book(self, book_path: str, book_name: str = 'New Book'):
        """
        Add a directory ful of files to the database
        :param book_path: The filepath to the directory that contains the files to add. e.g. lost_prince.txt
        :param book_name: The name of the book being added e.g. "Lost Prince"
        :return:
        """

        logger.info(f"autoanki: Adding book from [{book_path}]")

        # Clean the book
        if not self.book_cleaner.clean(book_path):
            logger.warning("autoanki: Unable to clean book [" + book_name + "].")
            return

        # Add the book to the database
        if not self.database_manager.add_book(book_path, book_name):
            logger.warning("Unable to add [" + book_name + "] to database.")
            return

        logger.info("autoanki: Added [" + book_path + "].")

    def complete_unfinished_definitions(self):
        """
        autoanki contains an internal definitions' table that is scraped from the internet. As words are added to
        autoanki, their definitions must be found. This function passively finds definitions and adds them to the table
        :return: None
        """

        # TODO Make progress par for unfinished records
        logger.info("Checking for records...")
        self.database_manager.cursor.execute("SELECT word FROM dictionary WHERE definition IS NULL")
        response_rows = self.database_manager.cursor.fetchall()
        while len(response_rows) > 0:
            self.database_manager.cursor.execute("SELECT word FROM dictionary WHERE definition IS NULL")
            response_rows = self.database_manager.cursor.fetchall()
            if len(response_rows) > 0:
                logger.info("Adding " + str(len(response_rows)) + " rows to dictionary table")
                for row in response_rows:
                    word = row[0]

                    # TODO This is a bad way of doing it, but find word is returning all of the parameters to
                    #     add to the database
                    # TODO create a dictionary that gets words from a file, not the internet
                    params = self.dictionary.find_word(word)
                    self.database_manager.complete_definition(params)

            else:
                logger.info("No new rows to complete in dictionary table")
            # time.sleep(2)

    @staticmethod
    def is_database(db_path):
        return DatabaseManager.is_database(db_path)

    @staticmethod
    def create_autoanki_db(db_path: str):
        DatabaseManager.create_autoanki_db(db_path)

    def create_deck(self, deck_name: str, filepath: str):
        """
        Creates a deck file in the directory of the main file.
        :return:
        """
        # FEATURE Add more options for how the deck looks
        # FEATURE get files from only one book, not the whole database

        logger.info("Generating deck file [" + deck_name + ".apk ]")
        words = self.database_manager.get_all_completed_definitions()

        deck_path = self.deck_manager.generate_deck_file(words, deck_name, filepath)
        if deck_path is None:
            logger.warning("Was not able to create deck file for [", deck_name, "]")
        else:
            logger.info("Generated deck file [" + deck_path + "]")

    @property
    def book_list(self):
        """
        Get a list of the books in the database
        :return: List of book names
        """
        return self.database_manager.book_list

    @book_list.setter
    def book_list(self, value):
        pass

    @property
    def unfinished_entries(self):
        return self.database_manager.unfinished_definitions()

    @unfinished_entries.setter
    def unfinished_entries(self, value):
        pass


if __name__ == '__main__':

    aa = AutoAnki()
    print(aa.book_list)
